#include "../../std.h"
#include <iostream>
#include <fstream>

#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>
#endif

#include "../../misc/escaped.h"
/*
#include "../../misc/inttostring.h"
#include "../../misc/stringtoint.h"
#include "../../misc/log_time.h"
#include "../../config/environment.h"
#include "../../auth/auth.h"
*/

#include "../jobs.h"
#include "jobs_soap.nsmap"

class HTTP_JobControl_Handle {
 friend class HTTP_JobControl;
 private:
  plugin_init_t jobplugin_init;
  void* jobplugin_handle;
 public:
  HTTP_JobControl_Handle(const char* path);
  ~HTTP_JobControl_Handle(void);
};

HTTP_JobControl_Handle::HTTP_JobControl_Handle(const char* path):jobplugin_handle(NULL),jobplugin_init(NULL) {
  jobplugin_handle=dlopen(path,RTLD_NOW);
  if(!jobplugin_handle) {
    odlog(0)<<dlerror()<<endl;
    odlog(0)<<"Can't load plugin "<<path<<endl;
    return;
  };
  jobplugin_init=(plugin_init_t)dlsym(jobplugin_handle,"init");
  if(jobplugin_init == NULL) {
    olog<<"Plugin "<<path<<" is broken."<<endl;
    dlclose(jobplugin_handle); jobplugin_handle=NULL; return;
  };
}

HTTP_JobControl_Handle::~HTTP_JobControl_Handle(void) {
  if(jobplugin_handle) {
    dlclose(jobplugin_handle); jobplugin_handle=NULL; jobplugin_init=NULL;
  };
}

HTTP_JobControl::HTTP_JobControl(HTTPS_Connector *c,const char* uri,HTTP_JobControl_Handle& handle):HTTP_ServiceAdv(c),service_url(uri) {
  /* SOAP */
  soap_init();
  sp.namespaces=jobs_soap_namespaces;
  sp.user=this;
  // Initialize plugin's instance
  userspec.fill(c->identity());
  string jobplugin_config = "end\n";
#ifdef HAVE_SSTREAM
  stringstream cfile(jobplugin_config);
#else
  strstream cfile;
  cfile<<jobplugin_config;
#endif
  jobplugin=handle.jobplugin_init(cfile,userspec);
  if(jobplugin == NULL) {
    olog<<"Job control plugin failed to initialise." << endl;
    return;
  };
  if(jobplugin->acquire() != 1) {
    olog<<"Job plugin's acquire failed (should never happen)." << endl;
    delete jobplugin; jobplugin=NULL;
    return;
  };
}

HTTP_JobControl::~HTTP_JobControl(void) {
  if(jobplugin) if(jobplugin->release() == 0) {
    odlog(0)<<"Destructor with dlclose"<<endl;
    delete jobplugin; jobplugin=NULL;
  };
  soap_deinit();
}

HTTP_Service* jobs_service_creator(HTTPS_Connector& c,const char* uri,void* arg) {
  HTTP_JobControl_Handle& handle = *((HTTP_JobControl_Handle*)arg);
  HTTP_JobControl *h = new HTTP_JobControl(&c,uri,handle);
  return h;
}

bool jobs_service_configurator(istream& f,const char* uri,HTTP_Service_Properties &prop) {
  string jobplugin_path;
  for(;!f.eof();) {
    char buf[1024];
    istream_readline(f,buf,sizeof(buf));
    char* p = buf;
    for(;*p;p++) if(!isspace(*p)) break;
    char* command = p;
    if(*command == '#') continue;
    for(;*p;p++) if(isspace(*p)) break;
    int command_len = p-command;
    if(!command_len) continue;
    if((command_len==3) && (strncmp(command,"end",command_len)==0)) { break; }
    else if((command_len==8) && (strncmp(command,"jobplugin",command_len)==0)) {
      input_escaped_string(p,jobplugin_path);
    };
  };
  HTTP_JobControl_Handle* handle = new HTTP_JobControl_Handle(jobplugin_path.c_str());
  prop.arg=handle;
  prop.subtree=true;
  return true;
}

static char* soap_strdup_l(soap* sp,const char* s,int l = -1) {
  if(l<0) l=strlen(s);
  char* s_ = (char*)soap_malloc(sp,l+1);
  if(s_ == NULL) return NULL;
  memcpy(s_,s,l); s_[l]=0;
  return s_;
}

#ifdef OLD_GSOAP
#error gSOAP older than 2.5.2 is not supported.
#else
void HTTP_JobControl::soap_methods(void) {
  odlog(3)<<"soap_methods: tag: "<<sp.tag<<endl;
  if(!soap_match_tag(&sp,sp.tag,"nj:submit")) soap_serve_nj__submit(&sp);
  else sp.error = SOAP_NO_METHOD;
}
#endif

/*
 * HTTP methods
 */
HTTP_Error HTTP_JobControl::get(const char* uri,int &keep_alive) {
  if(!jobplugin) {
    return HTTP_NOT_IMPLEMENTED;
  };
  HTTP_Error err;
  if((err=parse_header(keep_alive)) != HTTP_OK) return err;
  if(failure_parsing || unsupported_option_passed) {
    keep_alive=0;
    return HTTP_FAILURE;
  };
  // obtain requested name
  const char* name = uri+service_url.length();
  if(*name != 0) name++; // skip slash
  odlog(2)<<"HTTP_JobControl::get: name: "<<name<<endl;
  unsigned int l = strlen(name);
  if((l == 0) || (name[l-1] == '/')) { // directory
    list<DirEntry> dir_list;
    if(jobplugin->readdir(name,dir_list,DirEntry::full_object_info) != 0) {
      if(send_error_response(*c,keep_alive,416,NULL,"Error 416") != 0) {
        odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
      };
      return HTTP_OK;
    };
    // Form a page
    string page="<HTML><HEAD>"; page+=name; page+="</HEAD><BODY>";
    for(list<DirEntry>::iterator d = dir_list.begin();d!=dir_list.end();++d) {
      page+=d->name; page+="<BR>";
    };
    page+="</BODY>";
    // So far always return full content
    err=send_header(keep_alive,0,page.length()-1,false,page.length(),HTTP_Time(-1),HTTP_Time(-1));
    if(err != HTTP_OK) { keep_alive=0; return HTTP_ERROR; };
  } else { // file
    DirEntry fileinfo;
    if(jobplugin->open(name,GRIDFTP_OPEN_RETRIEVE) != 0) {
      if(send_error_response(*c,keep_alive,416,NULL,"Error 416") != 0) {
        odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
      };
      return HTTP_OK;
    };
    string name_(name);
    if(jobplugin->checkfile(name_,fileinfo,DirEntry::full_object_info) != 0) {
      jobplugin->close(false);
      if(send_error_response(*c,keep_alive,416,NULL,"Error 416") != 0) {
        odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
      };
      return HTTP_OK;
    };
    // so far only one range is effective
    uint64_t range_start_ = 0;
    uint64_t range_end_ = (uint64_t)(-1);
    if(range_passed) {
      range_start_=range_start[0];
      range_end_=range_end[0];
      if(range_end_ <= range_start_) {
        odlog(1)<<"range_end <= range_start"<<endl;
        jobplugin->close(false);
        if(send_error_response(*c,keep_alive,416,NULL,"Error 416") != 0) {
          odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
        };
        return HTTP_OK;
      };
    };
    if(fileinfo.size && (range_start_ >= fileinfo.size)) {
      jobplugin->close(false);
      if(send_error_response(*c,keep_alive,416,NULL,"Error 416") != 0) {
        odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
      };
      return HTTP_OK;
    };
    // Simple solution: always return <= 1MB of first range
    unsigned char buf[1024*1024];
    unsigned long long int size = sizeof(buf);
    if(range_end_ < (range_start_+size)) size=range_end_-range_start_;
    if(jobplugin->read(buf,range_start_,&size) != 0) {
      jobplugin->close(false);
      if(send_error_response(*c,keep_alive,416,NULL,"Error 416") != 0) {
        odlog(3)<<"failed to send 416"<<endl; keep_alive=0; return HTTP_ERROR;
      };
      return HTTP_OK;
    };
    jobplugin->close(false);
    range_end_=range_start_+size;
    err=send_header(keep_alive,range_start_,range_end_,true,fileinfo.size,HTTP_Time(-1),HTTP_Time(-1));
    if(err != HTTP_OK) { keep_alive=0; return HTTP_ERROR; };
  };
  return HTTP_OK;
}

HTTP_Error HTTP_JobControl::put(const char* uri,int &keep_alive) {
  return HTTP_NOT_IMPLEMENTED;
}

HTTP_Error HTTP_JobControl::post(const char* uri,int &keep_alive) {
  enum HTTP_Error r = soap_post(uri,keep_alive);
  return r;
}

/*
 *  SOAP interface functions
 */
int nj__submit(struct soap *sp,char* description,nj__jobinfo& r) {
  HTTP_JobControl* it = (HTTP_JobControl*)(sp->user);
  struct tm t_;
  struct tm *t;

  r.state=NULL;
  r.rsl=NULL;
  r.failure_code=(int*)soap_malloc(sp,sizeof(int));
  if(r.failure_code) *r.failure_code=1;
  r.failure_reason="Not implemented";
  return SOAP_OK;
}

